#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/param.h>
#include <sys/mount.h>

// ****    ****
int		audio_fd;
unsigned int	all_bytes;

//  ,  ,   -  0
int open_dsp(char *dsp_name, int freq, int *ffreq, int open_mode) {
  int format = AFMT_S16_LE;
  int channels = 2;
  int ec = 0;
  int dsp_cap;

  all_bytes = 0;

  if ((audio_fd = open(dsp_name, open_mode, 0)) == -1) {
    perror(dsp_name);
    return -1;
  };

  if (open_mode == O_RDWR) {
    if (ioctl(audio_fd, SNDCTL_DSP_SETDUPLEX, 0) == -1) {
      perror("SNDCTL_DSP_SETDUPLEX");
      ec--;
    } else {
      if (ioctl(audio_fd, SNDCTL_DSP_GETCAPS, &dsp_cap) == -1) {
        perror("SNDCTL_DSP_GETCAPS");
	ec--;
      } else {
        if (! (dsp_cap && DSP_CAP_DUPLEX)) {
	  ec--;
        }; //  
      }; //  
    }; //   
  }; //   

  if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &format) == -1) {
    perror("SNDCTL_DSP_SETFMT");
    ec--;
  };
  if (format != AFMT_S16_LE) {
    perror("AFMT_SB16_LE not supported");
    ec--;
  };

  if (ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &channels) == -1) {
    perror("SNDCTL_DSP_CHANNELS");
    ec--;
  };
  if (channels != 2) {
    perror("Stereo not supported");
    ec--;
  };

  if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &freq) == -1) {
    perror("SNDCTL_DSP_SPEED");
    ec--;
  };
  *ffreq = freq;

  if (ec) close(audio_fd);
  return ec;
}

//  ,   -  0
int close_dsp() {
  return close(audio_fd);
}

//  
int write_dsp(void *buf, size_t len) {
  int ec;
  all_bytes += len * 4;
  ec = write(audio_fd, buf, len * 4);
  if (ec == -1) perror("Write to DSP");
  return ec;
}

//   
int get_rest() {
  int rest;
  if (ioctl(audio_fd, SNDCTL_DSP_GETODELAY, &rest) == -1) {
    return 0;
  } else {
    return rest;
  }
}

// ,     
unsigned int get_position() {
  int rest;
  if (ioctl(audio_fd, SNDCTL_DSP_GETODELAY, &rest) == -1) {
    return all_bytes / 4;
  } else {
    return (all_bytes - rest) / 4;
  }
}

//  
int read_dsp(void *buf, size_t len) {
  int ec;
  ec = read(audio_fd, buf, len * 4);
  if (ec == -1) perror("Read from DSP");
  return ec;
}

//    
void stop_dsp() {
  int fake;
  if (ioctl(audio_fd, SNDCTL_DSP_RESET, &fake) == -1) {
    perror("SNDCTL_DSP_RESET");
  }
}

//      ,     
void post_dsp() {
  int fake;
  if (ioctl(audio_fd, SNDCTL_DSP_POST, &fake) == -1) {
    perror("SNDCTL_DSP_POST");
  }
}

// ****   ****

void get_free_space(char *fname, int b2s, int *mb, int *sc) {
  struct statfs buf;
  unsigned long long t;

  if (statfs(fname, &buf)) {
    *mb = 0; *sc = 0;
  } else {
    t = (unsigned long long)buf.f_bsize * (unsigned long long)buf.f_bavail;
    *mb = t >> 20;
    if (b2s == 0) { *sc = 0; } else { *sc = t / b2s; };
  }
}
