// LERA: Wave Format
//  / wav-, ,  
unit lrwvfm;

interface
uses inferno;

const
  wvMinVal	= Low (Integer); // Max/Min  
  wvMaxVal	= High(Integer);

type
  // 
  wvtFilterIndex= (wvflNothing, wvflLf2Rg, wvflRg2Lf, wvflSm2Bt, wvfl180);
  wvtFilter	= procedure(var buf; const ln: Cardinal);

  //  
  wvtIndicator	= procedure(ready: Single);

var
// 
  wvRdata 	: Cardinal;	//  
  wvRfreq	: Word;		//  (sample/sec)  
  wvRchanels	: Byte;		//  
  wvRrate	: Byte;		//     
  wvRposB	: Cardinal;	//       
  wvRposS	: Cardinal;	//        

// 
  wvWfreq	: Word;		//  (sample/sec)  
  wvWchanels	: Byte;		//  
  wvWrate	: Byte;		//     

// 
  wvIdata	: Cardinal;
  wvIfreq	: Word;
  wvIchanels	: Byte;
  wvIrate	: Byte;

//   wav ,   , False=ok
function wvfmNewFile(const NewName, InsName: String): Boolean;

//     wvRposS  ,
//    s2 
procedure getSample(var s1, s2: Integer);

//      wvRposB  ,
//     ,
//          
procedure getSamples(var dst; const FilterIndex: wvtFilterIndex; const ln: Cardinal; var rln: Cardinal);

//      FL2     ,
//        
procedure putSamples(var src; const ln: Cardinal; var FL2: File);

//        Dst,  Ins <> '',      
procedure CopyToFile(const InsFileName, DstFileName: String; const Start1, Stop1, Start2, Stop2: Cardinal; const Ind: wvtIndicator);

//          , wvRate  wvChanels
procedure ChangeFormat(const DstFileName: String; const Start, Stop: Cardinal; const Rate, Chanels: Byte;
		       const Freq: Word; const Filter: wvtFilterIndex; const Ind: wvtIndicator);

//   WAV 
procedure putHeader(var FL: File);

//    
procedure BuffersStatistic(var src; const ln: Cardinal; var mxL, mxR: Integer);


	// *****      *****
//       
procedure wvFilterLf2Rg(var buf; const ln: Cardinal);

//       
procedure wvFilterRg2Lf(var buf; const ln: Cardinal);

//       ,     
procedure wvFilterSm2Bt(var buf; const ln: Cardinal);

//   
procedure wvFilter180(var buf; const ln: Cardinal);

//    (  :)) )
procedure wvFilterNothing(var buf; const ln: Cardinal);

const
  wvFilters	: array[wvtFilterIndex] of record
    Name	: PChar;
    Routine	: wvtFilter;
  end=(
    (Name: '';	Routine: @wvFilterNothing),
    (Name: '  ';	Routine: @wvFilterLf2Rg),
    (Name: '  ';	Routine: @wvFilterRg2Lf),
    (Name: ' ';	Routine: @wvFilterSm2Bt),
    (Name: ' ';	Routine: @wvFilter180)
  );

  wvRlen	: Cardinal = 0;	//   ()  
  wvIlen	: Cardinal = 0;	//   

implementation
{$ASMMODE att}
const
  wvFileBufSize	= 1024000-1;	//   - 1  

type
  tBuf		= record
    Data	: array[0..wvFileBufSize] of Byte;
    posB, posE	: Cardinal;	// posB -     @Data[0], posE -   ,   Data
    Invalid	: Boolean;
  end;

{$PACKRECORDS 1}
  tRIFFHeader	= record
    Sign	: array[0..3] of Char;
    Leng	: Cardinal;
  end;

  tWAVEHeader	= record
    Sign	: array[0..3] of Char;
  end;

  tFMT_Header	= record
    Sign	: array[0..3] of Char;
    Leng	: Cardinal;
    Form	: Word;
    Chan	: Word;
    SamplePerSec: Cardinal;
    AvgBytesPerSec:Cardinal;	// Channels * SamplesPerSec * BitsPerSample / 8
    BlockAlign	: Word;		// Channels * BitsPerSample / 8
    BitsPerSample:Word;
    //FormatSpecific:Reserved
  end;

  tDATAHeader	= record
    Sign	: array[0..3] of Char;
    Leng	: Cardinal;
  end;
{$PACKRECORDS Default}

var
  BufS, BufB,			// BufS -     , BufB -   
  BufW		: tBuf;		// BufW -    ,   
  FL		: File;		//     
  FLi		: File;		//   

  rRIFFHeader	: tRIFFHeader;	// Read (Base) File
  rWAVEHeader	: tWAVEHeader;
  rFMT_Header	: tFMT_Header;
  rDATAHeader	: tDATAHeader;

  wRIFFHeader	: tRIFFHeader;	// Write File
  wWAVEHeader	: tWAVEHeader;
  wFMT_Header	: tFMT_Header;
  wDATAHeader	: tDATAHeader;

  iRIFFHeader	: tRIFFHeader;	// Insert File
  iWAVEHeader	: tWAVEHeader;
  iFMT_Header	: tFMT_Header;
  iDATAHeader	: tDATAHeader;


//     
procedure ResetRO(var f: File; const FaultOnError: Boolean);
var pm: Byte; r: Integer;
begin
{$I-}
  pm := FileMode;
  FileMode := 0;
  Reset(f, 1);
  FileMode := pm;
  if FaultOnError then begin r := IOResult; if r <> 0 then RunError(r); end;
{$I+}
end;

//   WAV- FL,      
procedure ReadWaveHeader(var FL: File;
  var RIFFHeader: tRIFFHeader;
  var WAVEHeader: tWAVEHeader;
  var FMT_Header: tFMT_Header;
  var DATAHeader: tDATAHeader;
  var Chanels, Rate: Byte; var Freq: Word; var Len, Data: Cardinal);
begin
  ResetRO(FL, False);
  if IOResult <> 0 then begin
    Freq := 44100; Chanels := 2; Rate := 2; Data := 0; Len := 0; exit;
  end;

  if ParamCount >= 4 then begin
    val(ParamStr(2), Freq); val(ParamStr(3), Chanels); val(ParamStr(4), Rate);
    if Chanels in [1, 2] then else addlog('     .');
    if Rate in [1, 2] then else addlog('       .');
    Data := 0; Len := FileSize(FL) div Chanels div Rate;
    close(FL); if IOResult <> 0 then Len := 0;
    exit;
  end;

  BlockRead(FL, RIFFHeader, SizeOf(tRIFFHeader));
  with RIFFHeader do begin
    if Sign <> 'RIFF' then AddLog('RIFF- ');
    if Leng <> FileSize(FL)-FilePos(FL) then AddLog('RIFF-  ');
  end;

  BlockRead(FL, WAVEHeader, SizeOf(tWAVEHeader));
  with WAVEHeader do begin
    if Sign <> 'WAVE' then AddLog('WAVE- ');
  end;

  BlockRead(FL, FMT_Header, SizeOf(tFMT_Header));
  with FMT_Header do begin
    if Sign <> 'fmt ' then AddLog('fmt - ');
    seek(FL, FilePos(FL) - (SizeOf(tFMT_Header)-8) + Leng);
    if Form <> 1 then AddLog(' PCM ');
    Chanels := Chan;
    if (Chan <> 1) and (Chan <> 2) then AddLog('  ');
    Freq := SamplePerSec;
    case BitsPerSample of
      8: Rate := 1;
     16: Rate := 2;
     else AddLog('    ');
    end;
    if AvgBytesPerSec <> Chan*SamplePerSec * BitsPerSample div 8 then AddLog('      ');
    if (BlockAlign <> BitsPerSample * Chan div 8) then AddLog('  ');
  end;

  repeat
    BlockRead(FL, DATAHeader, SizeOf(tDATAHeader));
    Data := FilePos(FL);
    Seek(FL, FilePos(FL) + DATAHeader.Leng);
  until (DATAHeader.Sign = 'data') or eof(FL);
  if DATAHeader.Sign <> 'data' then AddLog('      "data"');

  Len := DATAHeader.Leng div FMT_Header.BlockAlign;
  close(FL);
end;


//   WAV-,       wvR*,  ,
//         
function wvfmNewFile(const NewName, InsName: String): Boolean;
begin
  Assign(FL, NewName);
  ReadWaveHeader(FL, 
    rRIFFHeader, rWAVEHeader, rFMT_Header, rDATAHeader, 
    wvRchanels, wvRrate, wvRfreq, wvRlen, wvRdata);

  BufS.Invalid := True;
  BufB.Invalid := True;

  wvfmNewFile := False;
  
  wvIlen := 0;
  if InsName <> '' then begin
    Assign(FLi, InsName);
    ReadWaveHeader(FLi, 
      iRIFFHeader, iWAVEHeader, iFMT_Header, iDATAHeader, 
      wvIchanels, wvIrate, wvIfreq, wvIlen, wvIdata);
    if (wvIchanels <> wvRchanels) or
       (wvIrate    <> wvRrate   ) or
       (wvIfreq    <> wvRfreq   ) then begin wvIlen := 0; wvfmNewFile := True end;
  end;
end;


//    BufB   fr  ln (fr, ln - )    dst
procedure UnPack16_2(var dst; const fr, ln: Cardinal);
begin move(BufB.Data[fr * 4], dst, ln * 4) end;

procedure UnPack16_1(var dst; const fr, ln: Cardinal);
begin
  asm
    cld
    movl $BufB.Data, %esi
    movl dst,	%edi
    movl fr,	%ebx
    shll $1,	%ebx
    addl %ebx,	%esi
    movl ln,	%ecx
  .L1:
    lodsw
    stosw
    stosw
    loop .L1
  end ['esi', 'edi', 'ebx', 'ecx'];
end;

procedure UnPack08_2(var dst; const fr, ln: Cardinal);
begin
  asm
    cld
    movl $BufB.Data, %esi
    movl dst,	%edi
    movl fr,	%ebx
    shll $1,	%ebx
    addl %ebx,	%esi
    movl ln,	%ecx
  .L1:
    lodsw
    movb %ah,	%dh
    movb %al,	%ah
    xorb %al,	%al

    subb $0x80, %ah
    stosw

    movb %dh,	%ah

    subb $0x80, %ah
    stosw

    loop .L1
  end ['esi', 'edi', 'eax', 'ebx', 'ecx', 'edx'];
end;

procedure UnPack08_1(var dst; const fr, ln: Cardinal);
begin
  asm
    cld
    movl $BufB.Data, %esi
    movl dst,	%edi
    addl fr,	%esi
    movl ln,	%ecx
  .L1:
    lodsb
    subb $0x80,	%al
    movb %al,	%ah
    xorb %al,	%al
    stosw
    stosw
    loop .L1
  end ['esi', 'edi', 'eax', 'ecx'];
end;


//    Buf  FL   wvPos
procedure RefreshCache(var Buf: tBuf; const wvPos: Cardinal);
var RealSize: LongInt;
begin with Buf do begin
  ResetRO(FL, True); Seek(FL, wvRdata + wvPos*wvRrate*wvRchanels);
  BlockRead(FL, Data, wvFileBufSize + 1, RealSize); Close(FL);
  posB := wvPos; posE := wvPos + (RealSize div wvRrate div wvRchanels) - 1;
  if posE >= wvRlen then posE := wvRlen - 1;
  Invalid := False;
end end;


//     BufS
procedure getSample(var s1, s2: Integer);
var from: Cardinal;
begin
  if (wvRposS < BufS.posB) or (wvRposS > BufS.posE) or BufS.Invalid then RefreshCache(BufS, wvRposS);
  from := (wvRposS - BufS.posB) * wvRrate * wvRchanels;
  asm
    cld
    movl $BufS.Data, %ebx
    addl from, %ebx
    cmpb $1, wvRrate
    jne .L1

    xorb %al, %al

    movb (%ebx), %ah
    subb $0x80, %ah

    movw %ax, %dx
    incl %ebx

    movb (%ebx), %ah
    subb $0x80, %ah
    jmp .L2

  .L1:
    movw (%ebx), %dx
    addl $2, %ebx
    movw (%ebx), %ax

  .L2:
    movl s1, %ebx
    movw %dx, (%ebx)
    movl s2, %ebx
    movw %ax, (%ebx)
  end ['eax', 'ebx', 'edx'];
  inc(wvRposS);
end;


//    ln     dst,  rln     
//       FilterIndex
procedure getSamples(var dst; const FilterIndex: wvtFilterIndex; const ln: Cardinal; var rln: Cardinal);
var from: Cardinal;
begin
  if (wvRposB < BufB.posB) or (wvRposB+ln > BufB.posE+1) or BufB.Invalid then RefreshCache(BufB, wvRposB);
  from := wvRposB - BufB.posB;
  rln := BufB.posE + 1 - wvRposB; if rln > ln then rln := ln;
  if wvRrate = 1 then begin	//  8 bit
    if wvRchanels = 1 then UnPack08_1(dst, from, rln) else UnPack08_2(dst, from, rln);
  end else begin		// 16 bit
    if wvRchanels = 1 then UnPack16_1(dst, from, rln) else UnPack16_2(dst, from, rln);
  end;
  inc(wvRposB, rln);
  wvFilters[FilterIndex].Routine(dst, rln);
end;


//    src  ln     BufW
procedure Pack16_2(var src; const ln: Cardinal);
begin move(src, BufW.Data, ln * 4) end;

procedure Pack16_1(var src; const ln: Cardinal);
begin
  asm
    cld
    movl ln, %ecx
    movl src, %esi
    movl $BufW.Data, %edi
  .L1:
    lodsw
    stosw
    incl %esi
    incl %esi
    loop .L1
  end ['eax', 'ecx', 'esi', 'edi'];
end;

procedure Pack08_2(var src; const ln: Cardinal);
begin
  asm
    cld
    movl ln, %ecx
    movl src, %esi
    movl $BufW.Data, %edi
  .L1:
    incl %esi
    lodsb
    subb $0x80, %al
    movb %al, %ah

    incl %esi
    lodsb
    subb $0x80, %al
    xchgb %al, %ah
    stosw
    loop .L1
  end ['eax', 'ecx', 'esi', 'edi'];
end;

procedure Pack08_1(var src; const ln: Cardinal);
begin
  asm
    cld
    movl ln, %ecx
    movl src, %esi
    movl $BufW.Data, %edi
    incl %esi
  .L1:
    lodsb
    subb $0x80, %al
    stosb
    addl $3, %esi
    loop .L1
  end ['eax', 'ecx', 'esi', 'edi'];
end;


//  ln    FL2    src
procedure putSamples(var src; const ln: Cardinal; var FL2: File);
begin
  if wvWrate = 1 then begin	//  8 bit
    if wvWchanels = 1 then Pack08_1(src, ln) else Pack08_2(src, ln);
  end else begin		// 16 bit
    if wvWchanels = 1 then Pack16_1(src, ln) else Pack16_2(src, ln);
  end;
  BlockWrite(FL2, BufW.Data, ln * wvWrate * wvWchanels);
end;


//    ,     wvW*
procedure WriteWaveHeader(var FL: File; const NumSamples: Cardinal);
var DataLen: Cardinal;
begin
  DataLen := NumSamples * wvWrate * wvWchanels;

  wRIFFHeader.Sign := 'RIFF';
  wRIFFHeader.Leng := SizeOf(tWAVEHeader) + SizeOf(tFMT_Header) + SizeOf(tDATAHeader) + DataLen;
  BlockWrite(FL, wRIFFHeader, SizeOf(tRIFFHeader));

  wWAVEHeader.Sign := 'WAVE';
  BlockWrite(FL, wWAVEHeader, SizeOf(tWAVEHeader));

  with wFMT_Header do begin
    Sign := 'fmt ';
    Leng := SizeOf(tFMT_Header) - SizeOf(Sign) - SizeOf(Leng);
    Form := 1;
    Chan := wvWchanels;
    SamplePerSec := wvWfreq;
    case wvWrate of
      1: BitsPerSample :=  8;
      2: BitsPerSample := 16;
    end;
    AvgBytesPerSec := Chan * SamplePerSec * BitsPerSample div 8;
    BlockAlign := BitsPerSample * Chan div 8;
  end;
  BlockWrite(FL, wFMT_Header, SizeOf(tFMT_Header));

  wDATAHeader.Sign := 'data';
  wDATAHeader.Leng := DataLen;
  BlockWrite(FL, wDATAHeader, SizeOf(tDATAHeader));
end;


//       ,     wvW*    
procedure putHeader(var FL: File);
const HeaderSize = SizeOf(tRIFFHeader) + SizeOf(tWAVEHeader) + SizeOf(tFMT_Header) + SizeOf(tDATAHeader);
begin
  seek(FL, 0);  {
  wRIFFHeader := rRIFFHeader;
  wWAVEHeader := rWAVEHeader;
  wFMT_Header := rFMT_Header;
  wDATAHeader := rDATAHeader; }
  if FileSize(FL) <= HeaderSize then
    WriteWaveHeader(FL, 0)
   else
    WriteWaveHeader(FL, (FileSize(FL)-HeaderSize) div wvWrate div wvWchanels);
end;


// Dst := Cur[1..1] + Ins + Cur[2..2]
procedure CopyToFile(const InsFileName, DstFileName: String; const Start1, Stop1, Start2, Stop2: Cardinal; const Ind: wvtIndicator);
var
  FLdst		: File;
  SummSamples	: Cardinal;

  procedure FrCopy(var FLsrc: File; const Data, Start, Stop: Longint);
  var Ost, Len: Cardinal;
  begin
    seek(FLsrc, Data + Start * wvRrate * wvRchanels);
    Ost := (Stop - Start) * wvRrate * wvRchanels;
    While Ost <> 0 do begin
      if Ost > wvFileBufSize+1 then Len := wvFileBufSize+1 else Len := Ost;
      BlockRead (FLsrc, BufB.Data, Len);
      BlockWrite(FLdst, BufB.Data, Len);
      Ind( (FileSize(FLdst) div wvRrate div wvRchanels) / SummSamples );
      dec(Ost, Len);
    end;
  end;

begin
  BufB.Invalid := True;
  SummSamples := (Stop1-Start1) + (Stop2-Start2) + wvIlen;

  ResetRO(FL, True);
  assign(FLdst, DstFileName); rewrite(FLdst, 1);
		        {
  wRIFFHeader := rRIFFHeader;
  wWAVEHeader := rWAVEHeader;
  wFMT_Header := rFMT_Header;
  wDATAHeader := rDATAHeader;
			}
  wvWchanels := wvRchanels; wvWrate := wvRrate; wvWfreq := wvRfreq;
  WriteWaveHeader(FLdst, SummSamples);

  FrCopy(FL, wvRdata, Start1, Stop1);
  if InsFileName <> '' then begin
    ResetRO(FLi, True);
    FrCopy(FLi, wvIdata, 0, wvIlen);
    close(FLi);
  end;
  FrCopy(FL, wvRdata, Start2, Stop2);

  close(FL); close(FLdst);
end;


// 
procedure wvFilter180(var buf; const ln: Cardinal);
begin
  asm
    movl buf, %edi
    movl ln, %ecx
  .L1:
    negw (%edi)
    addl $4, %edi
    loop .L1
  end ['ecx', 'edi'];
end;

procedure wvFilterLf2Rg(var buf; const ln: Cardinal);
begin
  asm
    cld
    movl buf, %esi
    movl buf, %edi
    addl $2, %edi
    movl ln, %ecx
  .L1:
    lodsw
    stosw
    addl $2, %esi
    addl $2, %edi
    loop .L1
  end ['eax', 'ecx', 'esi', 'edi'];
end;

procedure wvFilterRg2Lf(var buf; const ln: Cardinal);
begin
  asm
    cld
    movl buf, %esi
    movl buf, %edi
    addl $2, %esi
    movl ln, %ecx
  .L1:
    lodsw
    stosw
    addl $2, %esi
    addl $2, %edi
    loop .L1
  end ['eax', 'ecx', 'esi', 'edi'];
end;

procedure wvFilterSm2Bt(var buf; const ln: Cardinal);
begin
  asm
    cld
    movl buf, %esi
    movl buf, %edi
    movl ln, %ecx
  .L1:
    lodsw
    cwtl //	cwde
//	cwtd
    movl %eax, %edx
    lodsw
    cwtl
    addl %edx, %eax
    shrl $1, %eax
    stosw
    stosw
    loop .L1
  end ['eax', 'ecx', 'edx', 'esi', 'edi'];
end;

procedure wvFilterNothing(var buf; const ln: Cardinal);
begin end;


// Dst := (new format) Cur[1..1]
procedure ChangeFormat(const DstFileName: String; const Start, Stop: Cardinal; const Rate, Chanels: Byte;
		       const Freq: Word; const Filter: wvtFilterIndex; const Ind: wvtIndicator);
const
  intBufSize	= (wvFileBufSize+1) div 4; //    

var
  FLdst		: File;
  Len, cLen, Ost: Cardinal;

begin
  if Start = Stop then exit;
  wvRposB := Start;

  assign(FLdst, DstFileName); rewrite(FLdst, 1);
		{
  wRIFFHeader := rRIFFHeader;
  wWAVEHeader := rWAVEHeader;
  wFMT_Header := rFMT_Header;
  wDATAHeader := rDATAHeader;
		}
  wvWchanels := Chanels; wvWrate := Rate; wvWfreq := Freq;
  Ost := Stop - Start;
  WriteWaveHeader(FLdst, Ost);

  repeat
    if Ost > intBufSize then Len := intBufSize else Len := Ost;
    getSamples(BufS.Data, Filter, Len, cLen);
    putSamples(BufS.Data, Len, FLdst);

    if Len <> cLen then AddLog('ChangeFormat:    :(( ' + valtostrG(len) + ' ' + valtostrG(clen) + ' ' + valtostrG(wvRposB));

    Ind(1 - Ost / (Stop-Start));
    dec(Ost, Len);
  until Ost = 0;

  close(FLdst);
end;


//   
procedure BuffersStatistic(var src; const ln: Cardinal; var mxL, mxR: Integer);
label Max;
{
type t = record l, r: Integer end;
var p: ^t; i: Cardinal;
begin
  p := @src;
  mxL := 0; mxR := 0;
  for i:=1 to ln do with p^ do begin
    if abs(l) > mxL then mxL := abs(l);
    if abs(r) > mxR then mxR := abs(r);
    inc(p);
  end;
end;
}
var mL, mR: Integer;
begin
  asm
    cld
    xorl %eax, %eax
    movw %ax, mL
    movw %ax, mR
    movl ln, %ecx
    movl src, %esi

  .L1:
    lodsl
    push %eax

    movw mL, %bx
    call Max
    movw %bx, mL
    
    pop %eax
    shrl $16, %eax
    
    movw mR, %bx
    call Max
    movw %bx, mR
    
    loop .L1

    jmp .L3
  {------------ %ax - sample, %bx - max }
  Max:
    cmpw $0, %ax
    jge .L2
    negw %ax
  .L2:
    cmpw %bx, %ax
    cmovgw %ax, %bx
    ret
  {------------}
  .L3:
  end ['eax', 'ebx', 'ecx', 'esi'];
  mxL := mL; mxR := mR;
end;

end.