// LERA: Wave IO (   /dev/dsp,     )
unit lrwvio;

interface
uses Inferno, lrmixr, Alf, lrwvvw, Mikky, lrwvfm, X, XUtil, XLib, X11, leraHd, VGAColors, Ini;

type
  iotAutoCtrlLevel = (aclMaxsSearch, aclNone, aclOptimal);

const
  ioRecACL	: iotAutoCtrlLevel = aclNone;
  ioLoopACL	: iotAutoCtrlLevel = aclMaxsSearch;
  ioPlayFL	: wvtFilterIndex = wvflNothing;
  ioRecFL	: wvtFilterIndex = wvflNothing;
  ioLoopFL	: wvtFilterIndex = wvflNothing;

  ioACLNames	: array[iotAutoCtrlLevel] of String[16]=(
    ' ',
    '',
    ' '
  );

procedure  PlayFile(Start, Stop: Cardinal);
procedure ShowLevel;
procedure   RecFile(const s: String);

implementation
const
  Border		= 8;

  WndIOX		: LongInt = 720;	//  
  WndIOY		: LongInt = 450;
  WndIOW		: LongInt = 294;
  WndIOH		: LongInt = 300;

  GraphX		= Border;		//   
  GraphY		= 50;

  intCodecSamples	= 1024;
  FontName		: String = '-*-*-medium-r-*-*-18-*-*-*-*-*-koi8-*';
  MaxQueue		: LongInt = 100;
  MinRest		: LongInt = 10000;
  dsp_name		: AnsiString = '/dev/dsp';
  StateString		: String = '';

  type
    tBuf		= array[1..intCodecSamples] of LongInt;
    tLevels		= Array[1..10000] of TXPoint;

  var
    WndIO		: TWindow;
    Event		: TXEvent;
    GC			: TGC;
    Col 		: TColors;
    OldFullingTime	: TFullingTime;
    cntMaxL, cntMaxR	: Cardinal;
    LevelsL, LevelsR	: tLevels;
    RealFreq,
    CurrentGraphX,
    GraphW, GraphH,
    MaxGraphX, GraphY0,
    GraphXL, GraphXR	: LongInt;


procedure intCalcGraphSize;
begin
  GraphW := WndIOW - Border * 2; GraphH := WndIOH - Border - GraphY;
  if wvWchanels > 1 then MaxGraphX := GraphW div 2 - Border else MaxGraphX := GraphW;
  GraphXL := GraphX;
  GraphXR := GraphX + MaxGraphX + Border;
  GraphY0 := GraphH + GraphY;  
end;

procedure intDrawWnd;

  procedure DrawGraph(const x: LongInt; var Lev: tLevels; const cntMax: Cardinal);
  var s: String[31]; i: LongInt;
  begin if (CurrentGraphX > 0) and (CurrentGraphX < MaxGraphX) then begin
    XSetForeGround(prDisplay, GC, Col[1]);
    XDrawLine(prDisplay, WndIO, GC, x, GraphY0, x + MaxGraphX, GraphY0);
    XDrawLine(prDisplay, WndIO, GC, x, GraphY, x, GraphY0);

    XSetForeGround(prDisplay, GC, Col[11]);
    XDrawPoints(prDisplay, WndIO, GC, @Lev, CurrentGraphX, CoordModeOrigin);

    XSetForeGround(prDisplay, GC, Col[8]);
    for i := 1 to CurrentGraphX do XDrawLine(prDisplay, WndIO, GC, i + x, Lev[i].y + 1, i + x, GraphY0);

    s := ValToStrG(cntMax);
    if cntMax = 0 then XSetForeGround(prDisplay, GC, Col[8]) else XSetForeGround(prDisplay, GC, Col[12]);
    AlfOutTextXY(WndIO, GC, x + MaxGraphX - AlfWidth(s), GraphY0 - AlfHeight('-'), s);
  end end;

begin
//  XClearWindow(prDisplay, WndIO);

  AlfSetFont(GC, FontName);
  XSetForeGround(prDisplay, GC, Col[15]);
  AlfOutTextXY(WndIO, GC, Border, Border, StateString + '                                 ');
  jkl(WndIO, GC, ' : ' + ValToStrG(RealFreq) + ' /');

  if CurrentGraphX = 1 then begin
    XSetForeGround(prDisplay, GC, Col[0]);
    XFillRectangle(prDisplay, WndIO, GC, GraphX, GraphY, GraphW, GraphH);
  end;

  DrawGraph(GraphXL, LevelsL, cntMaxL);
  if wvWchanels > 1 then DrawGraph(GraphXR, LevelsR, cntMaxR);
end;


procedure intBackGround(var eEvent: TXEvent; const BlockEn: Boolean);
const PrevState: String = '-';
begin
  OldFullingTime(eEvent, BlockEn);
  if (StateString <> PrevState) or ((eEvent.EventType = Expose) and (eEvent.XExpose.Window = WndIO)) then begin
    if StateString = '' then 
      XUnMapWindow(prDisplay, WndIO)
    else begin
      XMapWindow(prDisplay, WndIO);
      intDrawWnd;
    end;
    PrevState := StateString;
  end;
  if eEvent.EventType = ConfigureNotify then with eEvent.XConfigure do if Window = WndIO then begin
    WndIOW := Width; WndIOH := Height;
    intCalcGraphSize;
  end;
end;


procedure intAutoRange(var buf; const ln: Cardinal; const acl: iotAutoCtrlLevel);
const
  rMaxLevel	= 32000;
  rMinLevel	= 20000;
  llMax		= 50;

  llCntL	: LongInt = 0;
  llCntR	: LongInt = 0;

var
  mxL, mxR		: Integer;

  function CalcLev(const mx: Integer; var llCnt: LongInt): ShortInt;
  begin
    //   :
    // - llMax    
    // -   +-1,  ,   
    CalcLev := 0;
    if mx > rMaxLevel then CalcLev := -1;
    if (acl = aclOptimal) and (mx < rMinLevel) then begin
      inc(llCnt);
      if llCnt > llMax then begin
        llCnt := 0; CalcLev := 1;
      end;
    end else llCnt := 0;
  end;

begin
  if CurrentGraphX = MaxGraphX then CurrentGraphX := 1 else inc(CurrentGraphX);

  BuffersStatistic(buf, ln, mxL, mxR);
  if mxL > rMaxLevel then inc(cntMaxL);
  if mxR > rMaxLevel then inc(cntMaxR);
  With LevelsL[CurrentGraphX] do begin x := CurrentGraphX + GraphXL; y := GraphY0 - round(mxL / wvMaxVal * GraphH) end;
  With LevelsR[CurrentGraphX] do begin x := CurrentGraphX + GraphXR; y := GraphY0 - round(mxR / wvMaxVal * GraphH) end;

  intDrawWnd;

  if acl <> aclNone then mixrChangeVolume(CalcLev(mxL, llCntL), CalcLev(mxR, llCntR));
end;


procedure  PlayFile(Start, Stop: Cardinal);
var
  DrPos		: Cardinal;
  rest		: LongInt;
  IsPause, IsStop, IsEob: Boolean;
  Buf		: tBuf;
  Len, RLen	: Cardinal;

  procedure GetDriverPosition;
  begin DrPos := get_position() + Start end;

begin
  if Start >= Stop then exit;
  CurrentGraphX := 0;
  wvvwNewUserPosFlag := False;
  repeat
    if open_dsp(@(dsp_name[1]), mixrDACf, RealFreq, O_WRONLY) <> 0 then exit;
    mixrDACf := RealFreq;
    wvRposB := Start;
    mixrNewUserFreq := False;
    repeat
      rest := get_rest div 4;
      StateString := ' [' + ValToStrG(rest) + ' ]';

      FullingTime(Event, False);
      GetDriverPosition;

      if (wvRposB < Stop) and ((XPending(prDisplay) <= MaxQueue) or (rest < MinRest)) then begin
        if intCodecSamples + wvRposB > Stop then Len := Stop - wvRposB else Len := intCodecSamples;
        getSamples(Buf, ioPlayFL, Len, RLen);
        write_dsp(Buf, RLen);
      end;

      IsPause := TestMKey(skLAlt);
      IsStop := (KeyInResult = ord('s')) or ((CurrentWindow = WndIO) and (Event.EventType = DeleteNotify));
      IsEob := wvRposB >= Stop;

      wvvwNewPos(DrPos);
    until IsPause or IsStop or IsEob or wvvwNewUserPosFlag or mixrNewUserFreq;
    GetDriverPosition;
    if IsPause or IsStop or wvvwNewUserPosFlag or mixrNewUserFreq then stop_dsp() else begin post_dsp(); DrPos := Stop end;

    if IsPause then begin
      StateString := ''; intDrawWnd;
      Repeat FullingTime(Event, True) Until not TestMKey(skLAlt);
      wvvwNewUserPosFlag := False; //          
    end;

    if DrPos >= wvRlen then DrPos := 0;
    wvvwNewPos(DrPos);

    close_dsp();
    KeyInResult := 0;
    if wvvwNewUserPosFlag then begin Start := wvvwNewUserPos; wvvwNewUserPosFlag := False end else Start := DrPos;
  until IsStop or IsEob;
  StateString := '';
end;


procedure   RecFile(const s: String);
var
  FLdst		: File;
  mb, sc	: LongInt;
  AllSamples	: Cardinal;
  Buf		: tBuf;
  IsStop, IsPause:Boolean;

begin
  assign(FLdst, s); rewrite(FLdst, 1); putHeader(FLdst);
  cntMaxL := 0; cntMaxR := 0; CurrentGraphX := MaxGraphX;
  AllSamples := 0;

  repeat
    if open_dsp(@(dsp_name[1]), mixrADCf, RealFreq, O_RDONLY) <> 0 then exit;
    mixrADCf := RealFreq;
    mixrNewUserFreq := False;
    repeat
      get_free_space(@(s[1]), RealFreq * wvWchanels * wvWchanels, mb, sc);
      StateString := ' [' + wvvwSec2Str(AllSamples / RealFreq) + ' / ' + wvvwSec2Str(sc) + ']';
      FullingTime(Event, False);

      read_dsp(Buf, intCodecSamples);
      inc(AllSamples, intCodecSamples);
      intAutoRange(Buf, intCodecSamples, ioRecACL);
      wvFilters[ioRecFL].Routine(Buf, intCodecSamples);
      putSamples(Buf, intCodecSamples, FLdst);

      IsPause := TestMKey(skLAlt);
      IsStop := (KeyInResult = ord('s')) or ((CurrentWindow = WndIO) and (Event.EventType = DeleteNotify));
    until IsStop or IsPause or mixrNewUserFreq;

    if IsPause then begin
      StateString := ''; intDrawWnd;
      Repeat FullingTime(Event, True) Until not TestMKey(skLAlt);
    end;
    close_dsp();
    KeyInResult := 0;
  until IsStop;

  putHeader(FLdst);
  close(FLdst);
  StateString := '';
end;


procedure ShowLevel;
var
  fr			: LongInt;
  IsStop, IsPause	: Boolean;
  Buf			: tBuf;

begin
  cntMaxL := 0; cntMaxR := 0; CurrentGraphX := MaxGraphX;
  repeat
    StateString := ' ';
    fr := (mixrADCf + mixrDACf) div 2;
    if open_dsp(@(dsp_name[1]), fr, RealFreq, O_RDWR) <> 0 then exit;
    mixrNewUserFreq := False;
    repeat
      FullingTime(Event, False);

      read_dsp(Buf, intCodecSamples);
      intAutoRange(Buf, intCodecSamples, ioLoopACL);
      wvFilters[ioLoopFL].Routine(Buf, intCodecSamples);
      write_dsp(Buf, intCodecSamples);

      IsPause := TestMKey(skLAlt);
      IsStop := (KeyInResult = ord('s')) or ((CurrentWindow = WndIO) and (Event.EventType = DeleteNotify));
    until IsStop or IsPause or mixrNewUserFreq;

    if IsPause then begin
      StateString := ''; intDrawWnd;
      Repeat FullingTime(Event, True) Until not TestMKey(skLAlt);
    end;
    close_dsp();
    KeyInResult := 0;
  until IsStop;
  StateString := '';
end;


procedure intBuildWindow;
begin
  NewRootWindow(WndIOX, WndIOY, WndIOW, WndIOH, 10, 10, '  OSS', '  OSS', WndIO);
  GC := XCreateGC(prDisplay, WndIO, 0, NIL);
  SetVGAPalette(WndIO, Col);
  intCalcGraphSize;
end;


begin
  IniOpen('IO');
  IniGet('IO X', WndIOX); IniGet('IO Width',  WndIOW);
  IniGet('IO Y', WndIOY); IniGet('IO Heigth', WndIOH);
  IniGet('FontName', FontName);
  IniGet('DSP filename', dsp_name);
  IniGet('Maximum Queue', MaxQueue);
  IniGet('Minimum Rest', MinRest);
  IniClose;

  intBuildWindow;

  OldFullingTime := FullingTime; FullingTime := @intBackGround;
end.
